package me.totalfreedom.discord;

import discord4j.common.util.Snowflake;
import discord4j.core.DiscordClientBuilder;
import discord4j.core.GatewayDiscordClient;
import discord4j.core.event.domain.interaction.ChatInputInteractionEvent;
import discord4j.core.object.entity.Guild;
import discord4j.core.object.entity.Member;
import discord4j.core.object.entity.Role;
import discord4j.core.object.entity.channel.TextChannel;
import me.totalfreedom.discord.command.HelpCommand;
import me.totalfreedom.discord.command.ListCommand;
import me.totalfreedom.discord.command.TPSCommand;
import me.totalfreedom.discord.handling.CommandHandler;
import me.totalfreedom.discord.util.SnowflakeEntry;
import me.totalfreedom.discord.util.TFM_Bridge;
import me.totalfreedom.totalfreedommod.admin.Admin;
import me.totalfreedom.totalfreedommod.config.ConfigEntry;
import me.totalfreedom.totalfreedommod.player.PlayerData;
import me.totalfreedom.totalfreedommod.rank.Rank;
import me.totalfreedom.totalfreedommod.util.FLog;
import net.dv8tion.jda.internal.utils.concurrent.CountingThreadFactory;
import org.apache.commons.lang.RandomStringUtils;
import org.bukkit.Bukkit;
import reactor.core.publisher.Mono;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ScheduledThreadPoolExecutor;

public class Bot
{
    private final GatewayDiscordClient client;
    private final TFM_Bridge tfm;
    private final HashMap<String, PlayerData> LINK_CODES = new HashMap<>();
    private ScheduledThreadPoolExecutor RATELIMIT_EXECUTOR;
    private Boolean enabled = false;


    public Bot()
    {
        //Creates the gateway client and connects to the gateway
        this.client = DiscordClientBuilder.create(ConfigEntry.DISCORD_TOKEN.getString())
                .build()
                .login()
                .block();

        this.tfm = new TFM_Bridge(this);
    }

    public void initialize() {
        if (client == null) throw new IllegalStateException();

        final CommandHandler handler = new CommandHandler(client.getRestClient());

        /* Call our code to handle creating/deleting/editing our global slash commands.
        We have to hard code our list of command files since iterating over a list of files in a resource directory
         is overly complicated for such a simple demo and requires handling for both IDE and .jar packaging.
         Using SpringBoot we can avoid all of this and use their resource pattern matcher to do this for us.
         */
        List<String> commands = List.of("greet.json", "ping.json");

        try
        {
            handler.registerCommands(commands);
            handler.registerCommand(new HelpCommand());
            handler.registerCommand(new ListCommand());
            handler.registerCommand(new TPSCommand());
        } catch (Exception e)
        {
            Bukkit.getLogger().severe("Error trying to register global slash commands.\n" + e.getMessage());
        }
        //Register our slash command listener
        client.on(ChatInputInteractionEvent.class, handler::handle)
                .then(client.onDisconnect())
                .block(); // We use .block() as there is not another non-daemon thread and the jvm would close otherwise.

        RATELIMIT_EXECUTOR = new ScheduledThreadPoolExecutor(5, new CountingThreadFactory(this::poolIdentifier, "RateLimit"));
        RATELIMIT_EXECUTOR.setRemoveOnCancelPolicy(true);
    }

    public String getCode(PlayerData playerData) {
        for (String code : LINK_CODES.keySet())
        {
            if (LINK_CODES.get(code).equals(playerData))
            {
                return code;
            }
        }
        return "";
    }

    public String generateCode(int size)
    {
        return RandomStringUtils.randomNumeric(size);
    }

    public String formatBotTag() {
        return client.getSelf()
                .blockOptional()
                .orElseThrow()
                .getUsername() + "#" + client.getSelf()
                .blockOptional()
                .orElseThrow()
                .getDiscriminator();
    }

    public boolean syncRoles(Admin admin, String discordID)
    {
        if (discordID == null)
        {
            return false;
        }

        Guild server = client.getGuildById(SnowflakeEntry.serverID)
                .blockOptional()
                .orElseThrow();

        Member member = server.getMemberById(Snowflake.of(discordID))
                .blockOptional()
                .orElseThrow();

        Role adminRole = server.getRoleById(SnowflakeEntry.adminRoleID)
                .blockOptional()
                .orElseThrow();

        Role senioradminRole = server.getRoleById(SnowflakeEntry.seniorRoleID)
                .blockOptional()
                .orElseThrow();

        if (!admin.isActive())
        {
            syncRolesActivityCheck(member, adminRole, senioradminRole);
            return true;
        }

        if (admin.getRank().equals(Rank.ADMIN))
        {
            syncRolesAdminAssignment(member, adminRole, senioradminRole);
            return true;
        }
        else if (admin.getRank().equals(Rank.SENIOR_ADMIN))
        {
            syncRolesSeniorAssignment(member, adminRole, senioradminRole);
            return true;
        }
        return false;
    }

    private void syncRolesAdminAssignment(Member member, Role adminRole, Role senioradminRole) {
        member.getRoles().doFirst(() -> {
            if (!member.getRoles().collectList().blockOptional().orElseThrow().contains(adminRole))
            {
                member.addRole(adminRole.getId()).block();
            }
        }).doOnEach(r -> {
            Role role = r.get();
            if (role == null) return;

            if (role.equals(senioradminRole))
            {
                member.removeRole(role.getId()).block();
            }
        }).subscribe();
    }

    private void syncRolesActivityCheck(Member member, Role adminRole, Role senioradminRole) {
        member.getRoles().doOnEach(r -> {
            Role role = r.get();
            if (role == null) return;

            if (role.equals(adminRole) || role.equals(senioradminRole))
            {
                member.removeRole(role.getId()).block();
            }
        }).subscribe();
    }

    private void syncRolesSeniorAssignment(Member member, Role adminRole, Role senioradminRole) {
        member.getRoles().doFirst(() -> {
            if (!member.getRoles().collectList().blockOptional().orElseThrow().contains(senioradminRole))
            {
                member.addRole(senioradminRole.getId()).block();
            }
        }).doOnEach(r -> {
            Role role = r.get();
            if (role == null) return;

            if (role.equals(adminRole))
            {
                member.removeRole(role.getId()).block();
            }
        }).subscribe();
    }


    private String poolIdentifier()
    {
        return "TFD4J";
    }

    public TFM_Bridge getTFM()
    {
        return tfm;
    }

    public Mono<Guild> getGuildById()
    {
        return client.getGuildById(SnowflakeEntry.serverID);
    }

    public GatewayDiscordClient getClient()
    {
        return client;
    }

    public Map<String, PlayerData> getLinkCodes()
    {
        return LINK_CODES;
    }

    public boolean shouldISendReport()
    {
        if (ConfigEntry.DISCORD_REPORT_CHANNEL_ID.getString().isEmpty())
        {
            return false;
        }

        if (ConfigEntry.DISCORD_SERVER_ID.getString().isEmpty())
        {
            FLog.severe("No Discord server ID was specified in the config, but there is a report channel ID.");
            return false;
        }

        Guild server = client.getGuildById(SnowflakeEntry.serverID).block();
        if (server == null)

        {
            FLog.severe("The Discord server ID specified is invalid, or the bot is not on the server.");
            return false;
        }

        TextChannel channel = server.getChannelById(SnowflakeEntry.reportChannelID)
                .ofType(TextChannel.class)
                .block();

        if (channel == null)
        {
            FLog.severe("The report channel ID specified in the config is invalid.");
            return false;
        }

        return true;
    }
}
